#include "chil_global.h"

#define atoui(c) atoi(atoui_char_p = (c)); if (atoui_char_p < 0) return -1
char *atoui_char_p;

void
CHILUsage(ChilboltonData *data)
{
  char *program_name = "{program}";

  if (data->argv) {
    program_name = data->argv[0];
  }

  CHILPrintf("Usage:\n"
	     "   %s file1.dat raster1-rasterN [file2.dat raster1-rasterN ...]"
	     " [options]\n"
	     "   %s -all file1.{dat|raw} [file2.{dat|raw} ...] [options]\n"
	     "   %s -date YYYYMMDD filepattern1 [filepattern2 ...] [options]\n",
	     program_name, program_name, program_name);
  return;
}

void
CHILOptions(ChilboltonData *data)
{
  CHILPrintf("\n"
             "General data processing options:\n"
	     "      -averays n  average n rays\n"
	     "     -avegates n  average n gates\n"
	     "      -verbose    display raster and file information\n"
	     "        -quiet    display no information except errors\n"
	     "      -version    output version information\n"
	     "          -day n  only accept data written on day n\n"
	     "  -rangeoffset f  perform range correction with offset of f metres\n"
	     "         -Zcal f  add f dB to raw Zh values\n"
	     "        -clean    remove erroneous lines one pixel thick\n"
	     "  -doubleclean    remove erroneous lines one or two pixels thick\n"
	     "    -maxfields n  retrieve only the first n fields\n"
	     "       -sumdir s  set fallback summary-file directory to s\n"
	     "                     (default is current directory)\n"
	     "\n"
	     "Processing options specific to scanning data:\n"
	     "         -scan n  get scan n\n"
	     "       -ZDRcal f  add f dB to raw ZDR values\n"
	     "       -LDRcal f  add f dB to raw LDR values\n"
	     "     -PHIDPcal f  add f degrees to raw PHIDP values\n"
	     "       -maxLDR f  remove all data (probably clutter) with LDR greater than f\n"
	     "                     (-10 is a good value)\n"
	     "        -polar fZH fZV fZHV\n"
	     "                  Subtract noise from Z, ZDR and LDR using noise-equivalent\n"
	     "                     reflectivities at 1 km in the H, V and cross-polar\n"
	     "                     channels of fZ, fZV and fZHV (dBZ) respectively\n"
	     "    -threshold f  set noise-elemination threshold to f\n"
	     "      -doppler f  reject noisy velocity and phi_dp data with threshold f\n"
	     "\n"
	     "Processing options specific to cloud data:\n"
	     "      -process    subtract noise (Zh must be present) using gates\n"
	     "                  from the top of the ray to characterise noise\n"
	     "   -oscillator    subtract noise using gates from the top of the ray\n"
	     "                     to characterise oscillator interference\n"
	     "   -noisegates n  use n gates to characterise noise/interference\n"
	     "  -noisefactor f  set noise-elimination factor to f\n"
	     "    -noclutter n f   suppress clutter by removing data when both v and sigma_v\n"
	     "                     are less than f in the lowest n gates\n"
	     "      -nobreak    don't break data up into days\n"
	     " -correctrange    perform range correction\n"
	     "      -v_floor f  set equivalent noise floor for v to f dB above that for Zh\n"
	     "     -sv_floor f  set equiv. noise floor for sigma_v to f dB above that for Zh\n"
	     "       -reject n f   reject `range-glitched' rays with a Zh that jumps by more\n"
	     "                     than greater than f in two gates beyond gate n\n"
	     "      -minelev f  skip rays with an elevation less than f degrees\n");

  return;
}


int
CHILInit(int argc, char **argv, ChilboltonData *data)
{
  int i = 1, j, datelen, status;

  data->firsttime = 1;
  data->scan = 0;
  data->avegates = data->averays = 1;
  data->process = data->clean = 0;
  data->file = 0;
  data->noisegates = 25;
  data->newrangeoffset = data->noclutter = data->warn_glitch = 0;
  data->polarprocess = data->dopplerprocess = 0;
  data->warn_cloudnoise = data->warn_polarnoise = 0;
  data->warn_clutterLdr = data->warn_oscillator = data->warn_minelev = 0;
  data->warn_confused = data->warn_dopplernoise = 0;
  data->nobreak = 0;
  data->filebuf = NULL;
  data->threshold = 1.19;
  data->dopplerthreshold = 400.0;
  data->iray = data->iblock = data->nblock = 0;
  data->noisefactor = 3.0;
  data->rangeoffset = 0.0;
  data->Zcal = data->ZDRcal = data->PHIDPcal = data->LDRcal = 0.0;
  data->nez_Zh = data->nez_Zv = data->nez_Zhv = -99;
  data->maxLDR = 101;
  data->optstart = 0;
  data->maxfields = 127;
  data->verbose = data->quiet = data->progress = 0;
  data->reqday = data->reqmonth = data->reqyear = 0;
  data->rangecorrection = 0;
  data->sigma_v_floor = data->v_floor = 0.0;
  data->reject = 0;
  data->reject_gate = 0;
  data->reject_val = 0;
  data->minelev = -1;
  data->noclutter_minv = 0.1;
  data->noclutter_maxgate = 15;

  data->argv = argv;
  data->argc = argc;

  /* Reset some things to avoid segfaults if CHILEnd is called... */
  data->chil_file.position = FILE_CLOSED;
  data->chil_file.file = NULL;
  data->ray.range = NULL;
  data->file_list.allnames = NULL;

  if (argc < 2) {
    return CHIL_NOTFOUND;
  }
  if (strcmp(argv[1], "-date") == 0) {
    /* The program must read all summary files to decide what rasters
       to load */
    if (argc < 3) {
      CHILPrintf("No date specified!");
      return CHIL_ERROR;
    }
    datelen=strlen(argv[2]);
    if (datelen==6) {
      i=sscanf(argv[2],"%02u%02u%02u",
	       &data->reqyear, &data->reqmonth, &data->reqday);
    }
    else if (datelen==8) {
      i=sscanf(argv[2],"%04u%02u%02u",
	       &data->reqyear, &data->reqmonth, &data->reqday);
      data->reqyear -= 1900;
    }
    if (i<3) {
      CHILPrintf("Can't understand date: should be in the form `yymmdd' or `yyyymmdd'\n");
      return CHIL_ERROR;
    }
    if (argc < 4) {
      CHILPrintf("Need at least one filename\n");
      return CHIL_ERROR;
    }

    /* Must look for "-sumdir" argument! */
    for (i = 4; i < argc; ++i) {
      if (strcmp(argv[i],"-sumdir") == 0 && i+1<argc) {
	CHILSecondarySumDir(argv[++i]);
      }
    }

    if (CHILGetFileList(data->reqday, data->reqmonth, data->reqyear,
			argc-3, argv+3,
			&data->file_list) < 1) {
      CHILPrintf("Can't find any data in these files from %04u%02u%02u\n",
		 CHILTrueYear(data->reqyear), data->reqmonth, data->reqday);
      return CHIL_NOTFOUND;
    }
    /* Find the first switch */
    for (j = 4; j < argc; j++) {
      if (*argv[j] == '-' && *(argv[j]+1) != '\0')
	break;
    }
  }
  else if (strcmp(argv[1], "-all") == 0) {
    /* We want all rasters in all the files specified on the command line */
    if (argc < 3 || *argv[2] == '-') {
      CHILPrintf("No files specified\n");
      return CHIL_ERROR;
    }
    data->file_list.allnames = argv+2;
    /* Count the files / find the first switch */
    for (j = 3; j < argc && *argv[j] != '-'; j++);
    data->file_list.n = j-2;
  }
  else {
    /* The first pairs of arguments are filenames and raster ranges.
       For non-cloud data one would usually expect a single file and raster */
    j = 1;
    data->file_list.n = 0;
    while (j < argc) {
      /* first the filename */
      data->file_list.name[data->file_list.n] = argv[j];

      if (*(data->file_list.name[data->file_list.n]) == '-')
	/* found a switch: end of file/raster pairs */
	break;
      /* then the raster range */
      if (++j >= argc) {
	CHILPrintf("No raster number(s) specified for file\n");
	return CHIL_ERROR;
      }
      if (*argv[j] == '-' && *(argv[j]+1) == '\0') {
	data->file_list.allrasters[data->file_list.n] = 1;
      }
      else {
	data->file_list.allrasters[data->file_list.n] = 0;
	i = sscanf(argv[j],"%hu-%hu",&data->file_list.first[data->file_list.n],
		   &data->file_list.last[data->file_list.n]);
	if (i == 0) {
	  CHILPrintf("Cannot understand `%s' as raster range: reading all rasters\n",
		     argv[j]);
	  data->file_list.allrasters[data->file_list.n] = 1;
	} else if (i == 1) {
	  data->file_list.last[data->file_list.n]
	    = data->file_list.first[data->file_list.n];
	} else if (data->file_list.last[data->file_list.n] 
		   < data->file_list.first[data->file_list.n]) {
	  CHILPrintf("Last raster number is less than first raster number!\n");
	  data->file_list.last[data->file_list.n] 
	    = data->file_list.first[data->file_list.n];
	}
      }
      j++;
      (data->file_list.n)++;
    }
  }
  data->optstart = j;
  if (j < argc) {
    /* we have some switches to parse */
    for (i = j; i < argc; i++) {
      if (strcmp(argv[i],"-averays") == 0 && i+1<argc) {
	data->averays = atoui(argv[++i]);
      } else if (strcmp(argv[i],"-verbose") == 0) {
	data->verbose = 1;
      } else if (strcmp(argv[i],"-version") == 0) {
	CHILPrintf("Version: " CHIL_VERSION "\n");
	return CHIL_NOTFOUND;
      } else if (strcmp(argv[i],"-quiet") == 0) {
	data->quiet = 1;
      } else if (strcmp(argv[i],"-process") == 0) {
	data->process = 1;
      } else if (strcmp(argv[i],"-oscillator") == 0) {
	data->process = 2;
      } else if (strcmp(argv[i],"-noisegates") == 0 && i+1<argc) {
	data->noisegates = atoui(argv[++i]);
      } else if (strcmp(argv[i],"-maxfields") == 0 && i+1<argc) {
	data->maxfields = atoui(argv[++i]);
      } else if (strcmp(argv[i],"-clean") == 0) {
	data->clean = 1;
      } else if (strcmp(argv[i],"-doubleclean") == 0) {
	data->clean = 2;
      } else if (strcmp(argv[i],"-noclutter") == 0 && i+2<argc) {
	data->noclutter = 1;
	data->noclutter_maxgate = atoui(argv[++i]);
	data->noclutter_minv = atof(argv[++i]);
      } else if (strcmp(argv[i],"-nobreak") == 0) {
	data->nobreak = 1;
      } else if (strcmp(argv[i],"-correctrange") == 0) {
	data->rangecorrection = 1;
      } else if (strcmp(argv[i],"-rangeoffset") == 0 && i+1<argc) {
	data->newrangeoffset = 1;
	data->rangecorrection = 1;
	data->rangeoffset = atof(argv[++i]);
      } else if (strcmp(argv[i],"-Zcal") == 0 && i+1<argc) {
	data->Zcal = atof(argv[++i]);
      } else if (strcmp(argv[i],"-ZDRcal") == 0 && i+1<argc) {
	data->ZDRcal = atof(argv[++i]);
      } else if (strcmp(argv[i],"-LDRcal") == 0 && i+1<argc) {
	data->LDRcal = atof(argv[++i]);
      } else if (strcmp(argv[i],"-PHIDPcal") == 0 && i+1<argc) {
	data->PHIDPcal = atof(argv[++i]);
      } else if (strcmp(argv[i],"-threshold") == 0 && i+1<argc) {
	data->threshold = atof(argv[++i]);
      } else if (strcmp(argv[i],"-maxLDR") == 0 && i+1<argc) {
	data->maxLDR = atof(argv[++i]);
      } else if (strcmp(argv[i],"-polar") == 0 && i+3<argc) {
	data->nez_Zh = atof(argv[++i]);
	data->nez_Zv = atof(argv[++i]);
	data->nez_Zhv = atof(argv[++i]);
	data->polarprocess = 1;
      } else if (strcmp(argv[i],"-doppler") == 0 && i+1<argc) {
	data->dopplerthreshold = atof(argv[++i]);
	data->dopplerprocess = 1;
      } else if (strcmp(argv[i],"-reject") == 0 && i+2<argc) {
	data->reject_gate = atof(argv[++i]);
	data->reject_val = atof(argv[++i]);
	data->reject = 1;
      } else if (strcmp(argv[i],"-sumdir") == 0 && i+1<argc) {
	CHILSecondarySumDir(argv[++i]);
      } else if (strcmp(argv[i],"-noisefactor") == 0 && i+1<argc) {
	data->noisefactor = atof(argv[++i]);
      } else if (strcmp(argv[i],"-v_floor") == 0 && i+1<argc) {
	data->v_floor = atof(argv[++i]);
      } else if (strcmp(argv[i],"-sv_floor") == 0 && i+1<argc) {
	data->sigma_v_floor = atof(argv[++i]);
      } else if (strcmp(argv[i],"-minelev") == 0 && i+1<argc) {
	data->minelev = atof(argv[++i]);
      } else if (strcmp(argv[i],"-avegates") == 0 && i+1<argc) {
	data->avegates = atoui(argv[++i]);
      } else if (strcmp(argv[i],"-scan") == 0 && i+1<argc) {
	data->scan = atoui(argv[++i]);
      } else if (strcmp(argv[i],"-day") == 0 && i+1<argc) {
	data->reqday = atoui(argv[++i]);
      }
    }
  }

  if (data->file_list.n == 0) {
    CHILPrintf("No files specified!\n");
    return CHIL_ERROR;
  }

  if ((data->file_list.n>1 
       || data->file_list.first[0] < data->file_list.last[0])
      && (!data->quiet) && (!data->verbose))
    data->progress=1;

  data->file = -1;
    
  data->chil_file.position = FILE_CLOSED;
  status = CHILAdvance(data);
  if (status == CHIL_OK || status == CHIL_ENDOFBLOCK) {
    return CHIL_OK;
  }
  else {
    return status;
  }
}

int
CHILAdvance(ChilboltonData *data)
{
  char *file_name = "requested file";
  int status = 0;
  char allrasters = 1;
  ChilboltonFile *chil_file = &(data->chil_file);
  RasterInfo *raster_info = &(data->raster_info);
  Ray *ray = &(data->ray);

  if (data->file >= 0) {
    if (data->file_list.allnames) {
      allrasters = 1;
    }
    else {
      allrasters = data->file_list.allrasters[data->file];
    }
    if (chil_file->iscan == 0 || data->scan > 0) {
      ++(data->raster);
    }
  }
  while (1) {
    if ((data->file < 0)
	|| (chil_file->position == FILE_CLOSED)
	|| (chil_file->position == END_OF_DATA)
	|| (!allrasters && data->raster > data->file_list.last[data->file])) {

      if (chil_file->position != FILE_CLOSED)
	CHILClose(chil_file);

      if (++(data->file) < data->file_list.n) {
	if (data->file_list.allnames) {
	  file_name = data->file_list.allnames[data->file];
	}
	else {
	  file_name = data->file_list.name[data->file];
	}
	if ((status = CHILOpen(file_name, chil_file))
	    != START_OF_FILE) {
	  CHILPrintf("Failed to read %s\n", file_name);
	  return CHIL_NOTFOUND;
	}
      }
      else {
	/*
	if (!data->quiet) { 
	  CHILPrintf("\n"); 
	} 
	*/
	//CHILPrintf("CHILAdvace returns CHIL_ENDOFDATA\n");
	return CHIL_ENDOFDATA;
      }

      if (data->file_list.allnames) {
	allrasters = 1;
      }
      else {
	allrasters = data->file_list.allrasters[data->file];
      }

      if (chil_file->summary_exists == 0 && !allrasters) {
	CHILClose(chil_file);
	CHILPrintf("Can't find summary file\n");
	return CHIL_NOTFOUND;
      }

      if (data->verbose)
	CHILPrintFileInformation(chil_file);

      if (data->progress)
	CHILPrintf("Opened %s\n", file_name);

      /*
      if (data->progress && !allrasters)
	CHILPrintf("   reading raster ");
      */

      if (!allrasters) {
	data->raster = data->file_list.first[data->file];
      }
    }

    //CHILPrintf("%d %d %d\n", allrasters, chil_file->iscan, data->scan);

    if (((!allrasters) && chil_file->iscan == 0) || (data->scan > 0)) {

      if (data->progress)
	CHILPrintf("Reading raster %u\n", data->raster);

      if (CHILSeekRaster(chil_file, data->raster) == CHIL_OK ) {
	if (data->scan > 1) {
	  if (CHILSkipScans(chil_file, raster_info, data->scan-1)
	      != CHIL_OK) {
	    CHILPrintf("Warning: Scan %hu not found in raster\n", data->scan);
	    return CHIL_NOTFOUND;
	  }
	}
	break;
      }
      else {
	CHILPrintf("Warning: Raster %u not found in %s\n", data->raster,
		   file_name);
	++(data->raster);
	continue;
      }
    }
    break;
  }

  /* Found a good scan! */
  status = CHILReadInformationRecord(chil_file, raster_info);
  if (status == END_OF_DATA && chil_file->position != END_OF_DATA) {
    /* Oh dear - don't know where we are in this file/raster (corrupt?)
       ... best to set the position to END_OF_DATA and so skip it */
    data->warn_confused = 1;
    chil_file->position = END_OF_DATA;
  }

  if (data->newrangeoffset)
    raster_info->range_offset = data->rangeoffset;

  if (data->verbose)
    CHILPrintRasterInformation(raster_info);

  if (!data->firsttime)
    CHILFreeRay(ray);

  CHILInitRay(raster_info, data->avegates, data->averays, ray);

  ray->v_floor = data->v_floor;
  ray->sigma_v_floor = data->sigma_v_floor;

  if (data->verbose)
    CHILPrintFieldInformation(ray);

  if (raster_info->cloud_data) {
    return CHIL_OK;
  }
  else {
    //CHILPrintf("CHILAdvance returns CHIL_ENDOFBLOCK\n");
    /*
    if (!data->quiet) { 
      CHILPrintf("\n"); 
    } 
    */
    return CHIL_ENDOFBLOCK;
  }
}


int
CHILFetchRay(ChilboltonData *data)
{
  static int last_day = 0;
  static int last_month = 0;
  static int last_year = -1;
  int status;
  ChilboltonFile *chil_file = &(data->chil_file);
  RasterInfo *raster_info = &(data->raster_info);
  Ray *ray = &(data->ray);

  //CHILPrintf("?");
  while (1) {
    if ((status = CHILReadRay(chil_file,raster_info,ray)) == CHIL_OK) {

      if (data->reqday == 0 || ray->day == data->reqday) {

	if (raster_info->cloud_data) {

	  if (ray->elevation < data->minelev) {
	    data->warn_minelev = 1;
	    continue;
	  }

	  if (data->reject) {
	    if (CHILIsRangeGlitched(ray, data->reject_gate, data->reject_val)
		== CHIL_ERROR) {
	      data->warn_glitch = 1;
	      continue;
	    }
	  }
	  if (data->process == 1) {
            if (CHILSubtractNoise(ray,data->noisegates,data->noisefactor)
		!= CHIL_OK)
	      data->warn_cloudnoise = 1;
	  }
	  else if (data->process == 2) {
            if (CHILSubtractOscillatorNoise(ray,data->noisegates,
					    data->noisefactor)
		!= CHIL_OK)
	      data->warn_oscillator = 1; 
	  }

	  if (data->noclutter)
            CHILSuppressClutter(ray, data->noclutter_maxgate,
				data->noclutter_minv);

	  if (data->clean == 1)
            CHILCleanRay(ray);
	  else if (data->clean == 2)
	    CHILDoubleCleanRay(ray);

	  if (data->rangecorrection)
            CHILRangeCorrection(raster_info,ray);

	  if (data->Zcal != 0.0) 
	    CHILCalibrateZh(ray,data->Zcal);

	  if (!data->nobreak && last_day > 0 
	      && (last_day != ray->day || last_month != ray->month 
		  || !(last_year == ray->year 
		       || last_year == ray->year + 100))) {
	    status = CHIL_ENDOFBLOCK;
	    
	    /*
	    if (!data->quiet) { 
	      CHILPrintf("\n"); 
	    } 
	    */
	  }
	  else {
	    status = CHIL_OK;
	    /*
	    if (!data->quiet) {
	      CHILPrintf(".");
	    }
	    */
	  }

	  last_day = ray->day;
	  last_month = ray->month;
	  last_year = ray->year;

	  //CHILPrintf("CHILFetchRay on cloud data, returns %d\n", status);
	  return status;

	}
	else /* scanning data */ {

	  if (data->polarprocess) {
            if (CHILSubtractNoisePolar(ray, data->nez_Zh, data->nez_Zv,
				       data->nez_Zhv,
				       data->threshold) != CHIL_OK)
	      data->warn_polarnoise = 1;
	  }
	  if (data->maxLDR < 100) {
            if (CHILSuppressClutterWithLdr(ray, data->maxLDR) != CHIL_OK)
	      data->warn_clutterLdr = 1;
	  }
	  if (data->dopplerprocess) {
	    if (CHILRemoveNoiseDoppler(ray, data->dopplerthreshold) != CHIL_OK)
	      data->warn_dopplernoise = 1;
	  }

	  if (data->clean == 1)
            CHILCleanRay(ray);
	  else if (data->clean == 2)
	    CHILDoubleCleanRay(ray);

	  if (data->rangecorrection) {
            CHILRangeCorrection(raster_info,ray);
	  }

	  if (data->Zcal != 0.0) 
	    CHILCalibrateZh(ray, data->Zcal);

	  if (data->ZDRcal != 0.0) 
	    CHILCalibrateField(ray, ZDR, data->ZDRcal);

	  if (data->LDRcal != 0.0) 
	    CHILCalibrateField(ray, LDR, data->LDRcal);

	  if (data->PHIDPcal != 0.0) 
	    CHILCalibrateField(ray, PHIDP, data->PHIDPcal);

	}

	/* Got a good ray and processed it */
	/*
	if (!data->quiet) { 
	  CHILPrintf("."); 
	} 
	*/
	return CHIL_OK;
      }

    }
    else /* CHILReadRay failed */ {
      if ((status = CHILAdvance(data)) != CHIL_OK) {
	//CHILPrintf("CHILFetchRay returns %d\n", status);
	return status;
      }
    }
  }
}

void
CHILEnd(ChilboltonData *data)
{
  if (data->warn_cloudnoise)
    CHILPrintf("Warning: Attempt to subtract noise failed\n");
  if (data->warn_polarnoise)
    CHILPrintf("Warning: Attempt to subtract noise from polarisation parameters failed\n");
  if (data->warn_dopplernoise)
    CHILPrintf("Warning: Attempt to remove noise from doppler parameters failed\n");
  if (data->warn_oscillator)
    CHILPrintf("Warning: Attempt to subtract oscillator noise failed\n");
  if (data->warn_clutterLdr)
    CHILPrintf("Warning: Attempt to suppress clutter with Ldr failed\n");
  if (data->warn_glitch)
    CHILPrintf("Warning: Some data was rejected because of range-glitching\n");
  if (data->warn_minelev)
    CHILPrintf("Warning: Some data was rejected because the radar elevation was less than %f degrees\n",
	       data->minelev);
  if (data->warn_confused)
    CHILPrintf("Warning: Confused by data file - corrupt?\n");

  CHILFreeRay(&(data->ray));
  if (data->chil_file.position != FILE_CLOSED) {
    CHILClose(&(data->chil_file));
  }
}

/*
int
CHILPrintOptions(FILE *file)
{
  int i, j;
  if (chil_argv == NULL)
    {
      return CHIL_NOTFOUND;
    }
  for (i = chil_optstart; i < chil_argc; i++)
    {
      char *c = chil_argv[i];
      for (j = 0; c[j] != '\0'; j++)
	{
	  if (c[j] <= ' ')
	    {
	      fprintf(file," \"%s\"", c);
	      break;
	    }
	}
      if (c[j] == '\0')
	{
	  fprintf(file," %s", c);
	}
    }
  return CHIL_OK;
}
*/

char *
CHILGetOptions(ChilboltonData *data)
{
  int i, j, len = 0;
  char *ans;

  if (data->argv == NULL)
    {
      return NULL;
    }
  for (i = data->optstart; i < data->argc; i++)
    {
      char *c = data->argv[i];
      char need_quotes = 0;
      for (j = 0; c[j] != '\0'; j++)
	{
	  if (c[j] <= ' ')
	    {
	      need_quotes = 1;
	    }
	}
      len += (1 + j + need_quotes * 2);
    }

  ans = CHILCalloc(len, sizeof(char));
  if (ans == NULL) {
    CHILPrintf("Out of memory\n");
    return NULL;
  }

  len = 0;
  for (i = data->optstart; i < data->argc; i++)
    {
      char *c = data->argv[i];
      char need_quotes = 0;
      for (j = 0; c[j] != '\0'; j++)
	{
	  if (c[j] <= ' ')
	    {
	      need_quotes = 1;
	    }
	}
      if (need_quotes) {
	sprintf(ans + len, "\"%s\"", c);
      }
      else {
	sprintf(ans + len, "%s", c);
      }
      len += (1 + j + need_quotes * 2);
      ans[len -1] = ' ';
    }
  ans[len-1] = '\0';
  return ans;
}
