#include "chil_global.h"

/*
 * removes pixels where velocity std is less than min_v (ground clutter)
 * and velocity is near zero.
 */
int
CHILSuppressClutter(Ray *ray, uword max_gate, float min_v)
{
  word i, j, iSV = -1, iV = -1;
  /* find which parameter is std of v, if available */
  if ((iSV = CHILFindField(ray, V_STD)) == -1) {
    return CHIL_NOTFOUND;
  }
  /* find which parameter is v, if available */
  if ((iV = CHILFindField(ray, V2)) == -1) {
    if ((iV = CHILFindField(ray, V1)) == -1) {
      return CHIL_NOTFOUND;
    }
  }

  if (max_gate > ray->ngates)
    max_gate = ray->ngates;

  /* Then remove (most) ground clutter */
  if (ray->field_type[iSV] == V_STD) {
    for (i = 0; i < max_gate; i++) {
      if (*(ray->data[iSV]+i) < min_v && *(ray->data[iSV]+i)
	  != chil_no_signal_value && *(ray->data[iV]+i) > - min_v
	  && *(ray->data[iV]+i) < min_v) {
	for (j=0;j<ray->ntotal;j++) {
	  *(ray->data[j]+i)=chil_no_signal_value;
	}
      }
    }
  }
  return CHIL_OK;
}

/*
 * Simple noise subtraction, using the far gates to get the noise level.
 * Noise is subtracted from Z only; the remaining fields are masked only.
 * If using CHILCleanRay too, call this function FIRST
 */
int
CHILSubtractNoise(Ray *ray,int noisegates,float factor)
{
  word iZ=0,i=0,j;
  double sum_noiseP=0.0,sum_noiseP2=0.0,noiseP=0.0,P,std_noiseP;
  float sfloor[MAX_FIELDS];

  if (noisegates <1 || noisegates > ray->ngates) {
    CHILPrintf("CHILSubtractNoise: %d is an inappropriate number of noise gates\n",noisegates);
    return CHIL_ERROR;
  }

  /* find which parameter is Z */
  while (ray->field_type[iZ] != ZH && iZ<ray->ntotal)
    iZ++;
  if (ray->field_type[iZ] != ZH) {
    CHILPrintf("CHILSubtractNoise: attempt to subtract noise when Zh not a parameter\n");
    return CHIL_ERROR;
  }

  ray->endgate = ray->ngates - noisegates;

  /* Set the rejection thresholds for each parameter relative to that for Z */
  /* in multiplicative units */
  for (j=0;j<ray->ntotal;j++) {
    if (ray->field_type[j] == V1 || ray->field_type[j] == V2) {
      sfloor[j]=pow(10,0.1*ray->v_floor);
    }
    else if (ray->field_type[j] == V_STD) {
      sfloor[j]=pow(10,0.1*ray->sigma_v_floor);
    }
    else
      sfloor[j]=1.0;
  }


  /* estimate the noise level from the top gates */
  if (ray->range_normalised) {
    for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP = (pow(10,0.1* (*(ray->data[iZ]+i))))
	/ (ray->range[i]*ray->range[i]);
      sum_noiseP += noiseP;
      sum_noiseP2 += (noiseP * noiseP);
    }
  }
  else {
    for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP = (pow(10,0.1* (*(ray->data[iZ]+i))));
      sum_noiseP += noiseP;
      sum_noiseP2 += (noiseP * noiseP);
    }
  }

  ray->nez=10.0*log10(noiseP = sum_noiseP / noisegates);
  std_noiseP = sqrt(sum_noiseP2/noisegates - noiseP * noiseP);

  /* Subtract noise from Zh and mask other parameters*/
  if (ray->range_normalised) {
    for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))))/(ray->range[i]*ray->range[i]);
      for (j=0;j<ray->ntotal;j++) {
	if (P < sfloor[j]*noiseP + factor*std_noiseP) {
	  (*(ray->data[j]+i))=chil_no_signal_value;
	}
      }
      if ((*(ray->data[iZ]+i)) != chil_no_signal_value) {
	(*(ray->data[iZ]+i)) 
	  = (float) (10.0*log10((P-noiseP)*ray->range[i]*ray->range[i]));
      }
    }
  }
  else {
    for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))));
      for (j=0;j<ray->ntotal;j++) {
	if (P < sfloor[j]*noiseP + factor*std_noiseP) {
	  (*(ray->data[j]+i))=chil_no_signal_value;
	}
      }
      if (*(ray->data[iZ]+i) != chil_no_signal_value) {
	(*(ray->data[iZ]+i))=(float) (10.0*log10(P-noiseP));
      }
    }
  }
  return CHIL_OK;
}

/*
 * More advanced noise subtraction. 
 * Assumes there is interference from some kind of oscillator and uses 
 * noisegates gates from the end of the ray to characterise it.
 * Noise is subtracted from Z only; the remaining fields are masked only.
 * If using CHILCleanRay too, call this function FIRST
 */
int
CHILSubtractOscillatorNoise(Ray *ray,int noisegates,float factor)
{
  word iZ=0,i=0,j;
  double sum_diff_noiseP2=0.0,noiseP[1023],P,std_noiseP,sum_noiseP=0;
  float sfloor[MAX_FIELDS];

  if (noisegates <1 || noisegates > ray->ngates) {
    CHILPrintf("CHILSubtractNoise: %d is an inappropriate number of noise gates\n",noisegates);
    return CHIL_ERROR;
  }

  /* find which parameter is Z */
  while (ray->field_type[iZ] != ZH && iZ<ray->ntotal)
    iZ++;
  if (ray->field_type[iZ] != ZH) {
    CHILPrintf("CHILSubtractNoise: attempt to subtract noise when Zh not a parameter\n");
    return CHIL_ERROR;
  }

  ray->endgate = ray->ngates - noisegates;

  /* Set the rejection thresholds for each parameter relative to that for Z */
  /* in multiplicative units */
  for (j=0;j<ray->ntotal;j++) {
    if (ray->field_type[j] == V1 || ray->field_type[j] == V2) {
      sfloor[j]=pow(10,0.1*ray->v_floor);
    }
    else if (ray->field_type[j] == V_STD) {
      sfloor[j]=pow(10,0.1*ray->sigma_v_floor);
    }
    else
      sfloor[j]=1.0;
  }

  /* get the noise level from the top gates */
  if (ray->range_normalised) {
    for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP[i] = (pow(10,0.1* (*(ray->data[iZ]+i))))/(ray->range[i]*ray->range[i]);
      sum_noiseP += noiseP[i];
    }
  }
  else {
    for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP[i] = (pow(10,0.1* (*(ray->data[iZ]+i))));
      sum_noiseP += noiseP[i];
    }
  }

  ray->nez=10.*log10(sum_noiseP/noisegates);

  /* estimate the standard deviation from the difference between adjacent gates */
  for (i=ray->ngates-noisegates+1;i<ray->ngates;i++) {
    sum_diff_noiseP2 += (noiseP[i]-noiseP[i-1])*(noiseP[i]-noiseP[i-1]);
  }
  std_noiseP=sqrt(sum_diff_noiseP2/(noisegates-1));

  /* replicate the noise pattern down through the other gates */
  for (i=ray->ngates-noisegates-1;i>=0;i--) {
    noiseP[i]=noiseP[i+noisegates];
  }

  /* Subtract noise from Zh and mask other parameters*/
  if (ray->range_normalised) {
    for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))))/(ray->range[i]*ray->range[i]);
      for (j=0;j<ray->ntotal;j++) {
	if (P < sfloor[j]*noiseP[i] + factor*std_noiseP) {
	  (*(ray->data[j]+i))=chil_no_signal_value;
	}
      }
      if (*(ray->data[iZ]+i) != chil_no_signal_value) {
	(*(ray->data[iZ]+i))=(float) (10.0*log10((P-noiseP[i])*ray->range[i]*ray->range[i]));
      }
    }
  }
  else {
    for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))));
      for (j=0;j<ray->ntotal;j++) {
	if (P < sfloor[j]*noiseP[i] + factor*std_noiseP) {
	  (*(ray->data[j]+i))=chil_no_signal_value;
	}
      }
      if (*(ray->data[iZ]+i) != chil_no_signal_value) {
	(*(ray->data[iZ]+i))=(float) (10.0*log10(P-noiseP[i]));
      }
    }
  }
  return CHIL_OK;
}

/* Checks to see if the specified field (e.g. ZH) exceeds maxval at 
   the specified gate of the ray. If so, CHIL_ERROR is returned,
   otherwise CHIL_OK is returned. If the gate or field is out of 
   range, CHIL_NOTFOUND is returned. WARNING: be careful to use
   this function at the appropriate stage in the processing... */
int
CHILIsCorrupt(Ray *ray, uword gate, uchar field, float maxval)
{
  int ifield, val = 0;
  if (gate >= ray->ntotal || (ifield = CHILFindField(ray, field)) < 0) {
    return CHIL_NOTFOUND;
  }

  if ((val = *(ray->data[ifield]+gate)) > maxval 
      && val != chil_no_signal_value) {
    return CHIL_ERROR;
  }
  return CHIL_OK;
}

/* Looks for gate that has a Z more than maxjump greater than that two gates
   below; if so returns CHIL_ERROR, else returns CHIL_OK. If no Z, returns
   CHIL_NOTFOUND */
int
CHILIsRangeGlitched(Ray *ray, uword maxgate, float maxjump)
{
  int ifield = CHILFindField(ray, ZH);
  uword i;
  if (ifield < 0) {
    return CHIL_NOTFOUND;
  }
  for (i = maxgate; i < ray->ngates - 2; i++) {
    if (*(ray->data[ifield]+i) + maxjump < *(ray->data[ifield]+i+2)) {
      return CHIL_ERROR;
    }
  }
  return CHIL_OK;
}
