#include <iostream>
#include <cstring>
#include <cstdio>
#include <cmath>
#include <csignal>
#include "rplot.h"

extern "C" {
#include <netcdf.h>
}

#define DEBUG cerr << __FILE__ << ": line " << __LINE__ << "\n";

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    cerr << "NetCDF error on line " << __LINE__ << ": " \
      << nc_strerror(ncstatus) << "\n"; \
    exit(-1); \
} 

#ifndef DEFAULT_X_FONT_SERVER
#define DEFAULT_X_FONT_SERVER "unix/:7100"
#endif

enum FILE_FORMAT {
  FORMAT_POSTSCRIPT,
  FORMAT_PNG
};

enum PARAMETER {
  ZH,
  BETA
};

bool verbose = false, quiet = false;

char *
get_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  handle_error(nc_inq_atttype(ncid, varid, name, &type));
  if (type != NC_CHAR) {
    cerr << "Error: attribute " << name << " is not of type char\n";
    exit(-1);
  }
  handle_error(nc_inq_attlen(ncid, varid, name, &length));
  value = new char[length+1];
  if (!value) {
    cerr << "Out of memory\n";
    exit(2);
  }
  handle_error(nc_get_att_text(ncid, varid, name, value));
  value[length] = '\0';
  return value;
}

short int
get_short_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  short int value;

  handle_error(nc_inq_atttype(ncid, varid, name, &type));
  if (type != NC_SHORT) {
    cerr << "Error: attribute " << name << " is not of type short int\n";
    exit(-1);
  }
  handle_error(nc_inq_attlen(ncid, varid, name, &length));
  if (length > 1) {
    cerr << "Error: attribute " << name << " has more than 1 element\n";
  }
  handle_error(nc_get_att_short(ncid, varid, name, &value));
  return value;
}

float
get_float_attribute(int ncid, int varid, char *name)
{
  size_t length;
  float value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    if (!quiet) {
      cerr << "Warning: " << name << " attribute not found\n";
    }
    return -1.0e30;
  }
  if (length > 1) {
    cerr << "Error: " << name << " attribute has more than 1 element\n";
  }
  handle_error(nc_get_att_float(ncid, varid, name, &value));
  return value;
}

int
get_date_and_time(char *datestr, float time, int day, int month, int year)
{
  static char *months[] = {"Jan","Feb","Mar","Apr","May","June",
			   "July","Aug","Sept","Oct","Nov","Dec"};
  if (month < 1 || month > 12) {
    return 1;
  }
  else {
    int hour = (int) time;
    int minute = (int) (60.0 * (time - hour));
    sprintf(datestr, "%02d:%02d UTC, %d %s %d", hour, minute,
	    day, months[month-1], year);
  }
  return 0;
}

void
Usage(char **argv)
{
  exit(0);
}

#define MAX_FILES 2
#define MAX_OFFSET 20

int
main(int argc, char **argv)
{
  RFigure *figure;
  RAxes *axes;
  RFontLoader *font_loader;
  RColor blue(0, 0, 255), green(0, 255, 0), red(255, 0, 0), black(0, 0, 0), &col = green;
  FILE_FORMAT format=FORMAT_PNG;
  char *xfs = DEFAULT_X_FONT_SERVER;
  int width = 250, height = 230;
  int x_offset = 50, y_offset = 65;
  int fig_width, fig_height;
  int iarg, ifile, ivar = 0, nvar = 0;
  float maxoffset = 10.0/60.0;

  int ncid[MAX_FILES], rangeid, rangedimid;
  int timedimid[MAX_FILES], timeid[MAX_FILES], varid;
  size_t ngates, nrays[MAX_FILES];
  float lasttime[MAX_FILES];
  size_t start[2], count[2];
  int tmp;
  char tmpstr[256], datestr[256];
  short int year[MAX_FILES], month[MAX_FILES], day[MAX_FILES];

  float minheight = 0, maxheight = 12;
  int font_size = 12;
  float minvalue, maxvalue, interval;
  bool log_scale;
  float scale_offset;

  /* Parse switches */
  iarg=1;
  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg], "-verbose") == 0) {
      verbose = true;
    }
    if (strcmp(argv[iarg], "-png") == 0) {
      format = FORMAT_PNG;
    }
    else if ((strcmp(argv[iarg],"-ps") == 0) 
	     || (strcmp(argv[iarg],"-eps") == 0)) {
      format = FORMAT_POSTSCRIPT;
    }
    else if (strcmp(argv[iarg],"-heightrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minheight) < 1) {
	cerr << "Failed to read minimum height as float\n";
	exit(-1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxheight) < 1) {
	cerr << "Failed to read maximum height as float";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg], "-size") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%dx%d",
				   &width,&height) < 2) {
	cerr << "Failed to read size: should be in the form "
	  "<width>x<height>\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg], "-fontsize") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read font size\n";
	exit(-1);
      }
      font_size = atoi(argv[++iarg]);
      if (font_size < 1) {
	cerr << "Font size must be a positive integer\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg], "-maxoffset") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read maximum offset\n";
	exit(-1);
      }
      maxoffset = atof(argv[++iarg]);
    }
   else if (strcmp(argv[iarg],"-xfs") == 0) {
      if (argc > iarg+1) {
	xfs = argv[++iarg];
      }
      else {
	cerr << "X font server not specified";
	exit(-1);
      }
    }
    else {
      cerr << argv[iarg] << " not understood\n";
      Usage(argv);
    }
    ++iarg;
  }
  if (iarg > argc-1) {
    Usage(argv);
  }

  fig_width = width + x_offset + 20;
  fig_height = height + y_offset + 50;

  try {

    /* Contact font server */
    font_loader = new RFontLoader(xfs, format == FORMAT_PNG);
    font_loader->verbose(verbose);

    /* One plot to standard output */
    /* Initialise plot */
    if (format == FORMAT_POSTSCRIPT) {
      figure = new RFigure(fig_width, fig_height, POSTSCRIPT,
			   &cout, "Chilbolton cloud profile", 0.5);
    }
    else {
      figure = new RFigure(fig_width, fig_height, BITMAP_8BIT);
    }
    
    figure->clear();
    figure->axis(0,0,fig_width, fig_height);
    figure->font = font_loader->load(font_size,
				     string("helvetica"),
				     string("medium"),
				     string("r"),
				     string("adobe"));
    
    axes = new RAxes(figure, x_offset, y_offset, 
		     x_offset + width, y_offset + height);
    
    
    ivar = 0;
    nvar = 0;
    /* Loop through two file names on the command line */
    for (ifile = iarg; ifile < argc; ++ifile) {
      /* Get time dimension size */
      handle_error(nc_open(argv[ifile], NC_NOWRITE, &ncid[ivar]));
      handle_error(nc_inq_dimid(ncid[ivar], "time", &timedimid[ivar]));
      handle_error(nc_inq_dimlen(ncid[ivar], timedimid[ivar], &nrays[ivar]));
      /* Find time variable and last time in file */
      handle_error(nc_inq_varid(ncid[ivar], "time", &timeid[ivar]));
      start[0] = nrays[ivar]-1;
      handle_error(nc_get_var1_float(ncid[ivar], timeid[ivar], start,
				     &lasttime[ivar]));
      /* Get the date */
      day[ivar] = get_short_attribute(ncid[ivar], NC_GLOBAL, "day");
      month[ivar] = get_short_attribute(ncid[ivar], NC_GLOBAL, "month");
      year[ivar] = get_short_attribute(ncid[ivar], NC_GLOBAL, "year");

      ++ivar;
      if (ivar > 1) {
	break;
      }
    }
    nvar = ivar;

    if (nvar > 1) {
      if (lasttime[0] > lasttime[1] + maxoffset || day[0] > day[1]
	  || month[0] > month[1] || year[0] > year[1]) {
	handle_error(nc_close(ncid[1]));
	nvar = 1;   
      }
      else if (lasttime[1] > lasttime[0] + maxoffset || day[1] > day[0]
	  || month[1] > month[0] || year[1] > year[0]) {
	handle_error(nc_close(ncid[0]));
	ncid[0] = ncid[1];
	timeid[0] = timeid[1];
	timedimid[0] = timedimid[1];
	nrays[0] = nrays[1];
	lasttime[0] = lasttime[1];
	nvar = 1;
      }
    }

    for (ivar = 0; ivar < nvar; ++ivar) {
      float *range, *var;
      float missing_value;
      char *format = NULL;
      char varname[256] = "";

      handle_error(nc_inq_dimid(ncid[ivar], "range", &rangedimid));
      handle_error(nc_inq_dimlen(ncid[ivar], rangedimid, &ngates));

      /* Get the ranges */
      handle_error(nc_inq_varid(ncid[ivar], "range", &rangeid));
      handle_error(nc_inq_varndims(ncid[ivar], rangeid, &tmp));
      if (tmp != 1) {
	cerr << "Range variable is not a 1D vector\n";
	exit(-1);
      } 
      range = new float[ngates];
      start[0] = 0;
      count[0] = ngates;
      handle_error(nc_get_vara_float(ncid[ivar], rangeid, start, count,
				     range));

      /* Find Zh or beta */
      PARAMETER param = ZH;
      if (nc_inq_varid(ncid[ivar], "Zh", &varid) != NC_NOERR) {
	param = BETA;
	handle_error(nc_inq_varid(ncid[ivar], "beta", &varid));
      }

      handle_error(nc_get_att_float(ncid[ivar], varid, "missing_value",
				    &missing_value));

      if (param == ZH) {
	float freq;
	int freqid;

	log_scale = false;
	minvalue = -50;
	maxvalue = 20;
	interval = 10;
	scale_offset = 0;
	handle_error(nc_inq_varid(ncid[ivar], "frequency", &freqid));
	start[0] = 0;
	handle_error(nc_get_var1_float(ncid[ivar], freqid, start, &freq));
	sprintf(varname, "%d GHz radar reflectivity factor (dBZ)", (int) freq);
      }
      else {
	float wl;
	int wlid;

	log_scale = true;
	minvalue = 0;
	maxvalue = 3.5;
	interval = 1;
	scale_offset = -7;
	handle_error(nc_inq_varid(ncid[ivar], "wavelength", &wlid));
	start[0] = 0;
	handle_error(nc_get_var1_float(ncid[ivar], wlid, start, &wl));
	sprintf(varname, "%d nm lidar backscatter coefficient "
	  "(sr<SUP>-1</SUP> m<SUP>-1</SUP>)", (int) wl);
      }

      if (log_scale) {
	format = "10<SUP>%g";
      }

      var = new float[ngates];
      start[0] = nrays[ivar]-1;
      start[1] = 0;
      count[0] = 1;
      count[1] = ngates;
      handle_error(nc_get_vara_float(ncid[ivar], varid, start, count, var));

      axes->axis(minvalue+scale_offset, minheight, maxvalue+scale_offset,
		 maxheight);
      axes->clipbox_to_axis();
      for (register int i = 0; i < ngates; ++i) {
	if (var[i] <= missing_value) {
	  var[i] = minvalue + scale_offset;
	}
	else if (log_scale) {
	  var[i] = log10(var[i]) + scale_offset;
	}
	else {
	  var[i] = var[i] + scale_offset;
	}
      }

      if (ivar) {
	col = blue;
      }
      else {
	col = red;
      }

      for (register int i = 1; i < ngates; ++i) {
	if ((var[i] > minvalue+scale_offset
	     || var[i-1] > minvalue+scale_offset)
	    && range[i-1] >= minheight && range[i] <= maxheight) {
	  axes->plot(var[i-1], range[i-1], var[i], range[i], col, LINE_SOLID);
	}
      }
      axes->clipbox(false);
      delete[] var;
      delete[] range;
      axes->xticks(minvalue+scale_offset, interval, format,
		   DEFAULT_TICK_LENGTH, ivar, 1, 1);
      axes->pen_color(col);
      axes->xlabel(varname, ivar);
      axes->pen_color(black);
    }
    axes->pen_color(black);
    axes->yticks(minheight, (float) ((int) (1+(maxheight-minheight)/14)),
		 NULL, DEFAULT_TICK_LENGTH, 0, 1);
    axes->ylabel("Height (km)");
    axes->box();

    get_date_and_time(datestr, lasttime[0], day[0], month[0], year[0]);
    sprintf(tmpstr, "Profile at %s", datestr);
    figure->font = font_loader->load(font_size,
				     string("helvetica"),
				     string("bold"),
				     string("r"),
				     string("adobe"));
    figure->text(0.01, 0.0, tmpstr, ANCHOR_BOTTOMLEFT);
    if (format == FORMAT_PNG) {
      figure->write_png(&cout);
    }
    delete axes;
    delete figure;
    delete font_loader;
  }
  catch  (const char * error_message) {
    cerr << error_message << endl;
    return 1;
  }

  exit(0);
}
