#include <string.h>
#include <stdlib.h>
#include <netcdf.h>
#include "ceil.h"

#define ERROR(msg,status) fprintf(stderr, "Error: %s\n", msg); exit(status)
#define WARNING(msg) fprintf(stderr, "Warning: %s\n", msg)

#define FILE_PREFIX ""
#define FILE_SUFFIX "_lc.nc"
#define NO_SIGNAL -999.00

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}

void
add_attributes(int ncid, int fieldid, char *units, char *long_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
}

void
Usage(char **argv)
{
  fprintf(stderr, "Usage: cat file1.dat [file2.dat ...] | %s [options]\n"
	  "Options:\n"
	  "         -date YYYYMMDD output only data written on this date\n"
	  "      -averays n        average n rays\n"
	  "     -avegates n        average n gates\n"
	  "       -prefix dir      write netCDF files in specified directory\n"
	  "  -noisefactor f        set noise rejection threshold (default 4)\n"
	  "      -noclean          don't clean noisy rays\n"
	  "      -verbose          talk to me\n", argv[0]);
  exit(0);
}

int
main(int argc,char **argv)
{
  LCTime time;
  LCCloudBases cloudbases;
  LCBackscatterProfile profile_hr,profile;
  unsigned short int iarg=0,r_year=0,r_month=0,r_day=0;
  char bydate=0;
  char verbose=0, file_open=0, quiet=0, cleanray=1;
  int averays=1, avegates=1, rays_read, cloudgate;

  int rayid, gateid, wlid, altid, latid, lonid;
  int ncid, rangeid, timeid, betaid, stdnid;
  int basesid, base1id, base2id, base3id;
  int dimids[2], ngates=0;
  float tmp_float, heights[3];
  short int tmp_word;
  char aname[512], *prefix=".";
  char *tmp_name;
  size_t index=0, tmp_index;
  size_t start[2], count[2];

  float decimalhour, lastdecimalhour=-1, stdn, noisefactor=4.0;

  /* Process command-line arguments */

  while (++iarg < argc) {
    if (strcmp(argv[iarg],"-date") == 0 && iarg+1<argc) {
      bydate=1;
      if (sscanf(argv[++iarg],"%04hu%02hu%02hu",
		 &r_year,&r_month,&r_day) < 3) {
	ERROR("Failed to read date: should be in the form YYYYMMDD",1);
      }
    }
    else if (strcmp(argv[iarg],"-avegates") == 0 && iarg+1<argc) {
      avegates = atoi(argv[++iarg]);
      if (avegates < 1) {
	WARNING("Cannot average negative or zero gates");
	avegates = 1;
      }
    }
    else if (strcmp(argv[iarg],"-averays") == 0 && iarg+1<argc) {
      averays = atoi(argv[++iarg]);
      if (averays < 1) {
	WARNING("Cannot average negative or zero rays");
	averays = 1;
      }
    }
    else if (strcmp(argv[iarg],"-noisefactor") == 0 && iarg+1<argc) {
      noisefactor=atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-noclean") == 0) {
      cleanray=0;
    }
    else if (strcmp(argv[iarg],"-prefix") == 0 && iarg+1<argc) {
      prefix=argv[++iarg];
    }
    else if (strcmp(argv[iarg],"-verbose") == 0) {
      verbose=1;
      quiet=0;
    }
    else if (strcmp(argv[iarg],"-quiet") == 0) {
      quiet=1;
      verbose=0;
    }

    else {
      fprintf(stderr, "Error: `%s' not understood or not accompanied by enough subsequent arguments\n", argv[iarg]);
      Usage(argv);
    }
  }

  while (1) {
    rays_read = (LCGetAveragedRay(stdin,&time,&cloudbases,&profile_hr,
			       averays));
    if (rays_read) {
      /* PROCESS DATA */
      LCAverageRay(&profile_hr,&profile,avegates);
      if (cleanray) {
	LCCustomCleanRay(&cloudgate,&stdn,&profile,8,noisefactor);
      }
      decimalhour = ((float) time.hour) + ((float) time.minute)/60.0 +
	((float) time.second)/3600.0;
    }

    if (file_open && (!rays_read || (lastdecimalhour>decimalhour))) {
      /* CLOSE FILE */
      handle_error(nc_close(ncid));
      index = file_open = 0;
      if (bydate) {
	bydate = 2;
      }
      if (verbose) {
	fprintf(stderr, "\n");
      }
    }

    if (!rays_read || bydate == 2) {
      /* EXIT */
      exit(0);
    }

    if (!file_open && !(bydate && (r_year != time.year 
				  || r_month != time.month
				  || r_day != time.day))) {
      char *version = CHIL_VERSION;

      /* OPEN NEW FILE */
      sprintf(aname, 
	      "%s/" FILE_PREFIX "%04hu%02hu%02hu" FILE_SUFFIX,
	      prefix, time.year, time.month, time.day);
      /* Open netCDF file for writing */
      handle_error(nc_create(aname, NC_CLOBBER, &ncid));
      index = 0;
      if (!quiet)
	fprintf(stderr, "Writing %s\n", aname);

      /* Add some global attributes */
      tmp_name = "Vaisala CT75K lidar ceilometer";
      handle_error(nc_put_att_text(ncid, NC_GLOBAL, "system",
				   strlen(tmp_name), tmp_name));
 
      tmp_word = time.day;
      handle_error(nc_put_att_short(ncid, NC_GLOBAL, "day", NC_SHORT,
				    1, &(tmp_word)));

      tmp_word = time.month;
      handle_error(nc_put_att_short(ncid, NC_GLOBAL, "month", NC_SHORT,
				    1, &(tmp_word)));

      tmp_word = time.year;
      handle_error(nc_put_att_short(ncid, NC_GLOBAL, "year", NC_SHORT,
				    1, &(tmp_word)));

      handle_error(nc_put_att_text(ncid, NC_GLOBAL, "software_version",
				   strlen(version), version));
 
      /* Add time/ray dimension */
      handle_error(nc_def_dim(ncid, "time", NC_UNLIMITED, &rayid));

      /* Add range/gate dimension  */
      ngates = profile.ngates;
      handle_error(nc_def_dim(ncid, "range", ngates, &gateid));

      /* Add some scalar variables */
      handle_error(nc_def_var(ncid, "wavelength", NC_FLOAT, 0, dimids, 
			      &wlid));
      add_attributes(ncid, wlid, "nm", "Lidar wavelength");

      handle_error(nc_def_var(ncid, "latitude", NC_FLOAT, 0, dimids, 
			      &latid));
      add_attributes(ncid, latid, "degrees_north", NULL);

      handle_error(nc_def_var(ncid, "longitude", NC_FLOAT, 0, dimids, 
			      &lonid));
      add_attributes(ncid, lonid, "degrees_east", NULL);

      handle_error(nc_def_var(ncid, "altitude", NC_FLOAT, 0, dimids, 
			      &altid));
      add_attributes(ncid, altid, "m",
		     "Altitude of lidar above mean sea level");

      /* Add time coordinate variable */ 
      dimids[0] = rayid;
      handle_error(nc_def_var(ncid, "time", NC_FLOAT, 1, dimids,
			      &timeid));
      add_attributes(ncid, timeid, "hours UTC", "Decimal hours since midnight");

      /* Add range coordinate variable */
      dimids[0] = gateid;
      handle_error(nc_def_var(ncid, "range", NC_FLOAT, 1, dimids, &rangeid));
      add_attributes(ncid, rangeid, "km", "Range from lidar");

      /* Add cloud bases and stdn parameter */
      dimids[0] = rayid;
      handle_error(nc_def_var(ncid, "bases", NC_BYTE, 1, dimids,
			      &basesid));
      tmp_name = "Number of cloud bases";
      handle_error(nc_put_att_text(ncid, basesid, "long_name",
				   strlen(tmp_name), tmp_name));

      tmp_float = NO_SIGNAL;
      handle_error(nc_def_var(ncid, "base1", NC_FLOAT, 1, dimids,
			      &base1id));
      add_attributes(ncid, base1id, "km",
		     "Range of the first cloud base");
      handle_error(nc_put_att_float(ncid, base1id, "missing_value",
				    NC_FLOAT, 1, &tmp_float));

      handle_error(nc_def_var(ncid, "base2", NC_FLOAT, 1, dimids,
			      &base2id));
      add_attributes(ncid, base2id, "km",
		     "Range of the second cloud base");
      handle_error(nc_put_att_float(ncid, base2id, "missing_value",
				    NC_FLOAT, 1, &tmp_float));

      handle_error(nc_def_var(ncid, "base3", NC_FLOAT, 1, dimids,
			      &base3id));
      add_attributes(ncid, base3id, "km",
		     "Range of the third cloud base");
      handle_error(nc_put_att_float(ncid, base3id, "missing_value",
				    NC_FLOAT, 1, &tmp_float));

      if (cleanray) {
	handle_error(nc_def_var(ncid, "stdn", NC_FLOAT, 1, dimids,
				&stdnid));
	add_attributes(ncid, stdnid, "10^(-4) (sr km)^(-1)",
		       "Standard deviation of the backscatter coefficient "
		       "background noise");
      }

      /* Add backscatter coefficient */
      dimids[0] = rayid;
      dimids[1] = gateid;

      handle_error(nc_def_var(ncid, "beta", NC_SHORT, 2,
			      dimids, &betaid));
      add_attributes(ncid, betaid, "10^(-4) (sr km)^(-1)",
		     "Attenuated bacscatter coefficient");
      tmp_float = 0;
      handle_error(nc_put_att_float(ncid, betaid, "missing_value",
				    NC_SHORT, 1, &tmp_float));
      
      /* End define mode */
      handle_error(nc_enddef(ncid));

      /* Set the scalar variables */
      tmp_index = 0;
      tmp_float = LC_WAVELENGTH;
      handle_error(nc_put_var1_float(ncid, wlid, &tmp_index,
				     &tmp_float));

      tmp_float = CHIL_ALT;
      handle_error(nc_put_var1_float(ncid, altid, &tmp_index, &tmp_float));

      tmp_float = CHIL_LAT;
      handle_error(nc_put_var1_float(ncid, latid, &tmp_index, &tmp_float));

      tmp_float = CHIL_LON;
      handle_error(nc_put_var1_float(ncid, lonid, &tmp_index, &tmp_float));

      /* Set the range */
      for (tmp_index = 0; tmp_index < ngates; ++tmp_index) {
	tmp_float = profile.range[tmp_index] * 0.001;
	handle_error(nc_put_var1_float(ncid, rangeid,
				       &tmp_index, &tmp_float));
      }
      file_open = 1;
    }
   
    if (file_open) {
     /* WRITE RAY */
      handle_error(nc_put_var1_float(ncid, timeid, &index,
				     &decimalhour));
      handle_error(nc_put_var1_int(ncid, basesid, &index,
				   &(cloudbases.number)));

      heights[2] = heights[1] = heights[0] = NO_SIGNAL;
      if (cloudbases.number > 0) {
	heights[0] = cloudbases.height1 * 0.001;
	if (cloudbases.number > 1) {
	  heights[1] = cloudbases.height2 * 0.001;
	  if (cloudbases.number > 2) {
	    heights[2] = cloudbases.height3 * 0.001;
	  }
	}
      }

      handle_error(nc_put_var1_float(ncid, base1id, &index,
				   heights));
      handle_error(nc_put_var1_float(ncid, base2id, &index,
				   heights+1));
      handle_error(nc_put_var1_float(ncid, base3id, &index,
				   heights+2));

      if (cleanray) {
	handle_error(nc_put_var1_float(ncid, stdnid, &index,
				       &stdn));
      }
      start[0] = index;
      start[1] = 0;
      count[0] = 1;
      count[1] = ngates;
      handle_error(nc_put_vara_short(ncid, betaid, start,
				     count, profile.beta));
      ++index;
      if (verbose) {
	fprintf(stderr, ".");
      }
    }
    lastdecimalhour=decimalhour;
  }
}
